// Autor: Lukas

#version 330

// Diese drei Vektoren stammen als dessen out-Parameter aus dem vertex shader, um hier als Input verwendet zu werden.
in vec3 worldNormal;
in vec2 fragmentUV;
in vec3 worldPosition;
// Die Kameraposition wurde im vertex shader noch nicht verwendet.
uniform vec3 cameraLocation;
// Wichtig fr den genutzten Blinn-Phong-Shader:
uniform vec3 diffuse;
uniform vec3 specular;
uniform float shininess;
// Fr die folgenden Variablen siehe die ausfhrliche Erluterung in der main-Methode.
uniform int lichtquellenanzahl;
// Da in C++ gleich die Elementanzahl der Arrays angegeben werden muss, ich diese aber noch nicht wei,
// gebe ich interimsmig einfach 20 an (es werden wohl "in echt" viel weniger sein).
uniform vec3 lichtquellenposition[20];
uniform vec3 intensityDiffuse[20];
uniform vec3 intensitySpecular[20];
uniform float constantAttenuation[20];
uniform float linearAttenuation[20];
uniform float quadraticAttenuation[20];
// Auerhalb der Main-Methode definiert, da auch andere Methoden darauf zugreifen.
vec3 viewVektor;
vec3 myDiffuseColor;
vec3 normalizedWorldNormal;
// Die endgltige Pixelfarbe.
out vec4 fragColor;
// Verweist auf die zu verwendende Textur.
uniform sampler2D colorTexture;

//reflection
uniform sampler2D reflectionTex;

//Koordinaten fr die ShadowMap
in vec4 smTexCoord;
uniform sampler2DShadow shadowMap;

uniform vec4 clipPlane;

vec3 lightCalcDirection(int lichtquellenindex) {
    // Diese Methode liefert einen Richtungsvektor zurck, welcher von dem mit worldPosition
    // spezifizierten Punkt in die Richtung der Lichtquelle (spezifiziert mit lichtquellenindex) zeigt, deren Position in
    // der aktuellen Klasse in lichtquellenposition gespeichert ist. Will ich also einen Richtungsvektor
    // erzeugen, der von Punkt A ausgeht und nach Punkt B zeigt, so muss ich B - A rechnen:
    // A entspricht hier worldPosition und B lichtquellenposition[lichtquellenindex]. Dieser Richtungsvektor
	// wird auch gleich normalisiert und danach zurckgeliefert..
    vec3 richtungsvektor = normalize(lichtquellenposition[lichtquellenindex] - worldPosition);
    return richtungsvektor;
}


	

vec3 lightCalcIntensity(int lichtquellenindex, bool istDiffuse) {
    // Hier wird die Lichtintensitt (fr diffusen und specularen Anteil funktioniert die Methode gleichermaen;
	// es muss lediglich als boolean-Wert angegeben werden, worum es geht) errechnet, wie sie von worldPosition aus
	// wahrgenommen wird.
    // Hierfuer muss die Lichtschwaechung (bzw. Abschwaechung des Lichts, light attenuation)
    // zunaechst errechnet und danach die diffuseIntensity bzw. specularIntensity durch eben diese 
	// Lichtschwaechung dividiert werden.
    
	// D.h. an dem Ort, wo sich die Lichtquelle selbst befindet, wird auch ihre maximal
    // moegliche Intensitaet erreicht (die eben in den beiden intensity-Werten gespeichert ist); wird aber 
    // die Intensitaet an einem anderen Ort, welcher von dem Standort der Lichtquelle
    // weiter entfernt ist, gemessen, so wird diese geringer ausfallen. Kurz, je hoeher
    // die Lichtschwaechung, desto geringer ist die Lichtintensitaet an einem bestimmten
    // (durch Weltkoordinaten spezifizierten) Ort.
    
    // Die fuer die Errechnung der Lichtschwaechung A (attenuation) noetige Formel 
	// lautet gem des ECG-Wikis:
    // A = ac + d * al + (d^2) * aq
    // wobei ac den konstanten Abschwaechungskoeffizenten,
    // al den linearen Abschwaechungskoeffizenten,
    // aq den quadratischen Abschwaechungskoeffizenten 
    // und d die Entfernung zur Lichtquelle darstellt.
    
    // ac, al und aq sind in der aktuellen Klasse bereits als constantAttenuation,
    // linearAttenuation und quadraticAttenuation vorhanden; die Distanz zwischen
    // dem Punkt an der worldPosition und lichtquellenposition[lichtquellenindex], der Position der aktuell
	// betrachteten Lichtquelle, muss jedoch noch errechnet werden.
    
	// [Alte Version]
    // Dies ist allerdings rasch geschehen und laeuft aehnlich ab wie in lightCalcDirection():
    // Die erste Codezeile jener Methode kann fast uebernommen werden; auch hier
    // wird ein Vektor errechnet, welcher von worldPosition in Richtung lichtquellenposition[lichtquellenindex]
    // zeigt (in diesem Fall waere es aber eigentlich egal, ob er nun in diese Richtung
    // zeigt oder von lichtquellenposition[lichtquellenindex] nach worldPosition ausgerichtet ist, da es nur um die
    // Laenge geht). Normiert wird er diesmal allerdings nicht, sondern er erstreckt sich tatsaechlich
    // von dem einen Punkt zum anderen, steckt also die volle Distanz ab.
	//Vec3 richtungsvektor = lichtquellenposition[lichtquellenindex] - worldPosition;    	
    // Nun muss also nur noch die Laenge dieses Vektors ausgerechnet werden, wofuer es 
    // bereits eine praktische Methode fuer Vec3-Objekte gibt:
    //float distanz = length(richtungsvektor);
	// [/Alte Version]
	
	// Kurz und gut: Es muss einfach die Distanz zwischen lichtquellenposition[lichtquellenindex] und worldPosition,
	// also zwischen zwei Punkten, ausgerechnet werden, und genau hierfr gibt es bereits eine vorgefertigte Methode!
	float distanz = distance(lichtquellenposition[lichtquellenindex], worldPosition);
    // Nun liegen alle noetigen Werte fuer die Anwendung der obigen Formel vor:
    // Bei allen handelt es sich um float-Werte; auch distanz muss in dieser Form vorliegen,
    // da die obige Methode ebenfalls einen float-Wert zurueckliefert.
    // Achtung: Natrlich muss hier auch darauf geachtet werden, dass sich die attenuations auf die aktuelle
	// Lichtquelle beziehen!
    float lichtschwaechung = constantAttenuation[lichtquellenindex] + distanz * 
	linearAttenuation[lichtquellenindex] + (pow(distanz,  2)) * quadraticAttenuation[lichtquellenindex];
    // Nun muss nur noch die intensity durch die lichtschwaechung dividiert werden, um
    // die Lichtintensitaet am gewuenschten Punkt zu erhalten! Allerdings muss zuvor noch berprft werden, um welche
	// intensity es geht: intensityDiffuse oder intensitySpecular? Wenn istDiffuse true ist, dann ersteres, wenn nicht,
	// dann letzteres.

	vec3 abgeschwaechteLichtintensitaet;
    if (istDiffuse == true) {
        abgeschwaechteLichtintensitaet = intensityDiffuse[lichtquellenindex] / lichtschwaechung;
    }



	else {
        abgeschwaechteLichtintensitaet = intensitySpecular[lichtquellenindex] / lichtschwaechung;
    }


    	
	// Diese abgeschwchte Intensitt - welche es nun auch ist - muss nun nur noch zurckgeliefert werden.
    return abgeschwaechteLichtintensitaet;
}




vec4 blinn() 
{
    // Diese Implementierung eines Blinn-Phong-Shaders fut auf den Ausfhrungen von EVC-CG-Textblatt 8 zur Beleuchtung.
	// Zahlreiche Kommentare zum Prinzip sind in den Files meines ECG-Projekts des vorherigen Semesters enthalten und knnen
	// bei Bedarf hier eingefgt werden; es wurden lediglich Anpassungen an die andere Programmiersprache vorgenommen 
	// (besagtes Projekt war in Java verfasst).

	//shadowMap texture lookup
	float depth = textureProj(shadowMap, smTexCoord);
	//float inShadow = (depth < smTexCoord.z) ? 1.0 : 0.0;
	float inShadow = depth;

	//float inShadow = 1;
	//if(shadow < (smTexCoord.z)/smTexCoord.w )
	//}
	//	inShadow = 0;
	//}

	// zunchst werden R-, G- und B-Anteile der resultierenden Pixelfarbe initialisiert; jene werden mit jeder durchlaufenen
	// Lichtquelle immer wieder erhht.
	// Ambientes Licht hinzugefgt, indem als Anfangswert der Farbkanle Farbe * 0.1 ist.
	float lRot = myDiffuseColor.x * 0.1;
    float lGruen = myDiffuseColor.y * 0.1;
    float lBlau = myDiffuseColor.z * 0.1;
    // Es mssen alle vorhandenen Lichtquellen durchgegangen werden:
    for (int i=0; i<lichtquellenanzahl; i++) {
		//nur die erste Lichtquelle macht schatten:
		if(i > 0)
		{
			inShadow = 1.0;
		}

        // Der Richtungsvektor von unserem Punkt bei worldPosition zur Lichtquelle errechnet sich mit obiger Methode.
    	vec3 richtungZurLichtquelle = lightCalcDirection(i);
        // Errechnung von cosTheta (Theta = Winkel zwischen worldNormal und der Richtung zur Lichtquelle
    	float cosTheta = dot(normalizedWorldNormal, richtungZurLichtquelle);
        // cos(Theta) soll ein Intensitaetswert sein, und Intensitaeten < 0 machen keinen 
    	// Sinn! Also ueberpruefe ich einfach, ob ein stumpfer Winkel (ueber 90 Grad)
    	// vorliegt, sprich, der Cosinus von Theta negativ ist: In diesem Fall 
    	// setze ich cosTheta einfach 0 (bei spitzen Winkeln
    	// zwischen 0 und 90 Grad ist der Cosionus positiv und ich muss nichts aendern;
    	// bei einem rechten Winkel von 90 Grad wird das Ergebnis ohnehin automatisch
    	// gleich 0). 
    	if (cosTheta < 0) {
            cosTheta = 0;
        }

		//light intensity
		vec3 lightIntensityDiffuse = lightCalcIntensity(i, true);
		vec3 lightIntensitySpecular = lightCalcIntensity(i, false);

    		
    	// Nun muss nur noch die Halbierende ausgerechnet
    	// werden: Praktischerweise existiert fuer die Division durch die Laenge des
    	// Ergebnisvektors bereits eine Funktion. Da die Halbierende von der Richtung zur
    	// Lichtquelle abhaengt, muss sie in jedem Schleifendurchlauf neu errechnet werden.
		// Hierfr wird der Vektor vom Punkt zum Auge des Betrachters, sprich, von worldPosition zu
		// cameraLocation gebraucht: Dies ist der zuvor in der main errechnete viewVektor!
    	vec3 halbierende = normalize(richtungZurLichtquelle + viewVektor);
		// Siehe EVC-CG8, S.28, unten: Der Winkel zwischen Normalvektor n und h, der
    	// Halbierenden zwischen l und v, ist Sigma sehr aehnlich. Das skalare Produkt
    	// von n und h aehnelt also stark cosSigma, ist also "quasiCosSigma".
    	float quasiCosSigma = dot(normalizedWorldNormal, halbierende);
        // Ebenso wie cosTheta kann quasiCosSigma als Intensitaetswert aufgefasst
    	// werden, und Intensitaetswerte im negativen Bereich machen keinen Sinn!
    	// Deshalb wird dieser Wert auf 0 gesetzt, falls er negativ sein sollte.
    	if (quasiCosSigma < 0) {
            quasiCosSigma = 0;
        }


    		
    	// Nun wird die Pixelfarbe kalkuliert (siehe wiederum EVC-CG8), wobei jene
    	// natuerlich fuer R-, G- und B-Anteil
    	// separat errechnet werden muss (wobei n*l wie gehabt cos(Theta) entspricht
    	// und nun n * h als quasiCosSigma bezeichnet wird):

		// Siehe die entsprechende Methode: Mit "true" als zweiten Parameter errechnet lightsCalcIntensity()
		// die abgeschwchte diffuseIntensity, mit "false" die abgeschwchte specularIntensity.
    	lRot = lRot + myDiffuseColor.x * lightIntensityDiffuse.x * cosTheta * inShadow + 
    			specular.x * cosTheta * lightIntensitySpecular.x * 
    			pow(quasiCosSigma, shininess) * inShadow;
        // Da hier mit Skalaren gerechnet wird, ist es wichtig zu beachten, dass es sich
    	// auch bei n * h um ein skalares Produkt handelt, die dot-Methode oben verwendet wurde. 
		
		// Analog fr G und B:
		lGruen = lGruen + myDiffuseColor.y * lightIntensityDiffuse.y * cosTheta * inShadow + 
    			specular.y * cosTheta * lightIntensitySpecular.y * 
    			pow(quasiCosSigma, shininess) * inShadow;
        lBlau = lBlau + myDiffuseColor.z * lightIntensityDiffuse.z * cosTheta * inShadow + 
    			specular.z * cosTheta * lightIntensitySpecular.z * 
    			pow(quasiCosSigma, shininess) * inShadow;
    }


    	
    // Der Rest erfolgt wie gehabt - nun muss lediglich noch die Pixelfarbe fr den aktuellen Punkt 
    // in RGB-Form, also als Vec3, zurueckgeliefert werden:
    vec3 rgbPixelfarbeAktuell = vec3(lRot, lGruen, lBlau);
    // Da als Rckgabewert aber ein vec4 erwartet wird, wird einfach noch ein Einser als homogene Komponente ergnzt.
	return vec4(rgbPixelfarbeAktuell, 1);
}



// Die Farbe wird zugewiesen.
void main()
{
	float clipPos = dot (worldPosition, clipPlane.xyz) + clipPlane.w;
    if (clipPos < 0.0) {
        discard;
    }
	
    // Es wird die - bereits interpolierte - Farbe an der Stelle (u,v) der Textur aufgerufen...
	vec3 textureColor = texture(colorTexture, fragmentUV).rgb;
    // ...und mit dem diffuse-Vektor multipliziert, um die endgltige diffuse Farbe herauszubekommen.
	myDiffuseColor = diffuse * textureColor;
    // In worldNormal und worldPosition sind bereits interpolierte Vertexposition und interpolierter Normalvektor
	// gespeichert, welche ich bald als Parameter brauchen werde.

	// Nun kann endlich der View-Vektor (welcher auch gleich normiert wird) berechnet werden!

	normalizedWorldNormal = normalize(worldNormal);
    viewVektor = normalize(cameraLocation - worldPosition);
    // Ich habe nun nahezu alle Parameter, um meine Blinn-Methode aufzurufen; allerdings brauche ich hier noch einige
	// Informationen zu den Lichtquellen: Diesmal ist es nicht so einfach wie in ECG, dass einfach ein lights[]-Array
	// im Shader vorhanden ist! Diesmal muss ich berprfen, welche Werte ich fr die Blinn-Berechnungen brauche, mir diese
	// von Blender bergeben lassen und als uniform-Inputs des fragment shaders speichern!

	// Dafr brauche ich folgendes:
	// I. Eine Angabe darber, wie viele Lichtquellen vorhanden sind.
	// II. Den Richtungsvektor vom Punkt zur Lichtquelle (fr jede Lichtquelle; erhalten via Methode lightCalcDirection())
	// III. Die tatschliche (abgeschwchte) Intensitt jeder Lichtquelle (lightCalcIntensity())

	// Fr diese Vorhaben brauche ich jedoch (auer fr I.) fr jede Lichtquelle noch weitere Informationen:
	// zu II.: Die Position der Lichtquelle
	// zustzlich fr III.: Die ursprngliche Intensitt der Lichtquelle; die Abschwchungen constant, linear und quadratic 
	// Attenuation

	// WICHTIG: Offenbar erhalten wir von Blender keinen "intensity"-Vektor fr die Lichtquelle, 
	// sondern einen fr "diffuse" und einen fr "specular" der Lichtquelle; also nenne ich diese "intensityDiffuse"
	// und "intensitySpecular".

	// Also erwarte ich, von Blender (bzw. von einer zustzlichen Klasse, welche die Informationen aus Blender herausliest
	// und mir selbige in der gewnschten Gestalt in den fragment shader liefert) folgende inputs (jeweils als uniform, da
	// fr jeden Punkt eines Objekts, eigentlich sogar fr die ganze Szene gleichbleibend):

	// int lichtquellenanzahl
	// vec3 lichtquellenposition (in Weltkoordinaten)
	// vec3 intensityDiffuse (die eigentliche/ursprngliche Intensitt der Lichtquelle, welche nach ihrer Abschwchung
	// mit dem diffuse-Vektor der Objektoberflche multipliziert wird)
	// vec3 intensitySpecular (die eigentliche/ursprngliche Intensitt der Lichtquelle, welche nach ihrer Abschwchung
	// mit dem specular-Vektor der Objektoberflche multipliziert wird)
	// float constantAttenuation
	// float linearAttenuation
	// float quadraticAttenuation

	// Daraus errechnen kann ich mir dann jeweils
	// vec3 richtungZurLichtquelle und 
	// vec3 abgeschwaechteIntensitaet (die tatschliche Intensitt unter Bercksichtigung der attenuations)

	// Abgesehen von der Lichtquellenanzahl muss jeder dieser Werte separat fr jede Lichtquelle gespeichert werden, weshalb
	// ich auf folgende Lsung mit Arrays der gleichen Namen wie der Einzelwerte oben zurckgreife: 
	// lichtquellenposition[0] gibt die Position der ersten Lichtquelle an,
	// lichtquellenposition[1] jene der zweiten etc.; fr die anderen Arrays gilt es analog, so gibt. z.B. 
	// quadraticAttenuation[3] eben jene Abschwchung fr die vierte Lichtquelle an.

	// Mit all diesen Informationen (welche nicht als Parameter angegeben werden mssen, da sie sich alle in Variablen 
	// dieser Klassen befinden) wird nun noch diese Methode aufgerufen, welche die endgltige Pixelfarbe, welche den
	// einzigen out-Parameter darstellt, bestimmt.
	fragColor = blinn();
    // Mittelalte Version
	// Nun wird endlich die Texturfarbe errechnet:
	// fragColor ist, wie oben definiert, ein Vec4, da darin R-, G- B- und Alpha-Kanal gespeichert sind.
	//fragColor = texture(colorTexture, fragmentUV);

	// Alte Version
	//vec3 textureColor = texture(colorTexture, fragmentUV).rgb;
	// Alte Demo-Version aus der Vorlesung.
	//vec3 textureColor = (vec3(fragmentUV, 0.0f) + 1 ) / 2;
	// Angabe von RGB und Alphabkanal (letzterer auf 1.0 gesetzt)
	// fragColor = vec4(textureColor, 1.0);

	//vec2 resolution;
	//resolution.x = 800;
	//resolution.y = 600;

	//zur fragColor die Spiegelung dazuaddieren:
	//if(gl_ClipDistance[0] < 0)
	//{
	//fragColor.rgb = texture(reflectionTex, gl_FragCoord.xy * resolution.xy).rgb;
	//}
}