// Autor: Lukas

// Das Prinzip orientiert sich an meinen Shader-Implementierungen aus ECG des vorherigen Semesters.
#version 330

// analog zu den Vorlesungsfolien; Position, Normale und uv-Koordinaten werden fr jeden Vertex aufgerufen; diese Variablen
// werden also bei jedem Aufruf berschrieben.
in vec3 position;
in vec3 normal;
in vec2 uv;
// uniform: Kann einmal definiert werden und gilt dann solange, bis es berschrieben wird; praktisch fr die folgenden
// Werte, welche jeweils fr alle Vertices des aktuellen Objekts gelten.

// Diese Matrizen werden auerhalb errechnet und dann von hier aus angesprochen.
uniform mat4 modelMatrix;
uniform mat4 viewMatrix;
uniform mat4 projectionMatrix;
uniform mat4 projectionViewModelMatrix;
uniform mat4 viewModelMatrix;
// Alles, was ich hier im VertexShader via "out" ausgebe, kommt interpoliert (in Bezug auf die drei Vertices des Dreiecks) 
// im FragmentShader an.

//shadowMap vorberechnete Matrix
uniform mat4 depthProjectionViewModelMatrix;

// Die Normale in Weltkoordinaten
out vec3 worldNormal;
// Diese uv-Koordinaten geben an, welcher Teil der Textur an der Stelle dieses Vertex sichtbar sein soll.
out vec2 fragmentUV;
// Die Vertexposition in Weltkoordinaten.
out vec3 worldPosition;

//ausgabe fr die ShadowMap
out vec4 smTexCoord;

//water clipping plane
//vorzeichen umdrehen um die "andere Seite" zu rendern
uniform vec4 u_plane0 = vec4 (0, 1, 0, -5);

void main() {

    // (1) Weiterreichen der Texturkoordinaten:
	// Die uv-Koordinaten werden weitergereicht, zwischenzeitlich interpoliert und in den fragment shader bernommen.
	fragmentUV = uv;
    // (2) Die Vertex-Position von Modellkoordinaten in normalised projection coordinates (npc) umwandeln: Nachdem dann 
	// auch noch (automatisch von OpenGL getan) geclippt wird, liegt der Wert in normalized device coordinates (ndc) vor.
	// Der Fragment Shader wird dann automatisch fr diese ndc-Position aufgerufen (auch darum kmmert sich OpenGL).
	// Vorgefertigte Methode gl_Position: Dies wird automatisch zum Output!
	// Hier wird einfach nur die Position weitergereicht. Vec3 --> Vec 4: Homogene Komponente 1 wird ergnzt.
	// Anm.: Erst NACH dem Clipping darf gl_Position homogenisiert werden - das liegt nicht in der Macht des vertex shaders!
	// Deswegen wird es hier nicht getan.
	//gl_Position = projectionMatrix * viewMatrix * modelMatrix * vec4(position, 1);

	//besser mit der vorberechneten projectionViewModelMatrix
	gl_Position = projectionViewModelMatrix * vec4(position, 1);

	//fr den lookup in der ShadowMap
	smTexCoord = depthProjectionViewModelMatrix * vec4(position, 1);

    // (3) Vertex-Position und Normalvektor in Weltkoordinaten transformieren (was spter noch fr einige Berechnungen
	// gebraucht wird).
	// Ersteres Vorgehen wird mithilfe einer Multiplikation gelst - hnlich (2), aber simpler: Es wird einfach die 
	// Position mittels der modelMatrix in Weltkoordinaten bergefhrt.
	vec4 worldPosition4D = modelMatrix * vec4(position, 1);
    // Da ich die Position in Weltkoordinaten aber als 3D-Vektor haben will, homogenisiere ich sie (Division durch homogene
	// Komponente w) und wandle sie in einen 3D-Vektor um (wurde oben bereits als Output-Parameter vec3 worldPosition
	// definiert).
	worldPosition = worldPosition4D.xyz / worldPosition4D.w;
    // Schwieriger ist es dagegen, eine Normale in Weltkoordinaten zu transferieren; hier kann nicht einfach wie bei einem 
	// Punkt die Model Matrix verwendet werden! 
	// Es ist etwas komplizierter - die folgenden Ausfhrungen und Codezeilen orientieren sich an einer meiner Abgaben
	// fr ECG im vorherigen Semester, wo dieser Fall ebenfalls thematisiert wurde.

	// Siehe den "Transforming Normal Vectors"-Artikel aus dem ECG-Wiki:
    	// Wird eine Oberflaeche per Transformationsmatrix M transformiert, kann ein Tangentialvektor t,
    	// welcher besagte Oberflaeche in einem Punkt beruehrt, ebenfalls einfach in gleichartiger
    	// Weise mit M multipliziert werden, damit dieser neu errechnete, upgedatete 
    	// Tangentialvektor die nun veraenderte Oberflaeche tangiert.
    	// Weniger trivial ist es dagegen bei der korrekten Transformation der Oberflaechennormale n
    	// besagter Oberflaeche: Einfach nur diesen Normalvektor mit M zu multiplizieren stellt nicht
    	// sicher, dass dabei wirklich ein korrekter Normalvektor fuer die veraenderte Oberflaeche
    	// herauskommt. Gesucht ist deshalb eine andere Transformationsmatrix N, welche, wenn
    	// mit n multipliziert, den korrekten neuen Normalvektor der veraenderten Oberflaeche
    	// ergibt.
    	// Aeuesserst praktisch wird nun, dass fuer eine Oberflaeche generell gilt, dass
    	// Oberflaechennormale und Tangentialvektor (bzw. Oberflaechennormale und Tangentialebene, auf welcher sich die
    	// Tangentialvektoren befinden) zueinander in rechtem Winkel stehen und damit die Formel
    	// n[transponiert]*t=0
    	// gilt.
    	// Fuer die mittels M veraenderte Oberflaeche gilt nun Folgendes:
    	// t[M] = M * t ...der neue Tangentialvektor, welcher mit den gegebenen
    	// Werten leicht ausgerechnet werden kann.
    	// n[N] = N * n ...die neue Oberflaechennormale, welche gesucht ist und nach der Anwendung von N 
		// als Matrix zur Normalvektortransformation azf den gegebenen Modellkoordinaten-Normalvektor entsteht.
    	// Dabei muss gelten n[N][transponiert] * t = 0 (siehe oben)
    	// Nun folgen gemaess dem Artikel von der Angabe einige "algebraic tricks":
    	// Die Einheitsmatrix I (besitzt ausser den 1er-Werten in der Hauptdiagonale nur
    	// Nullen als Elemente) hat ja die Eigenschaft, dass M invers * M = I.
    	// Eine Multiplikation einer Matrix wie M mit der Einheitsmatrix ergibt, wenn die
    	// unten stehenden Voraussetzungen erfuellt sind, wieder die unveraenderte Matrix M:
    	// I[m]*M=M*I[n]=M, wobei M eine mxn-Matrix ist und "I[m]" meint "Die Einheitsmatrix
    	// muss m Zeilen haben" und mit "I[n]" gemeint ist "Die Einheitsmatrix muss n 
    	// Spalten haben".
    	// Unter diesen Voraussetzungen kann in das skalare Produkt n[transponiert]*t=0
    	// problemlos eine Einheitsmatrix "hineingeschmuggelt" werden, welche danach dann
    	// durch M invers * M ersetzt wird: n[transponiert]*t=n[transponiert]*I*t=
    	// n[transponiert]*M invers*M*t=0.
    	// Oder geklammert, vereinfacht und besser als skalares Produkt erkennbar angeschrieben:
    	// (n[transponiert]*M invers)*(M*t)=(n[transponiert]*M invers)*t[M]=0
    	// da ja, siehe oben, t[M]=M*t gilt.
    	// Diese Formel kann fuer alle Tangentialvektoren, welche sich auf der Tangentialebene
    	// befinden, angewandet werden und gibt an, dass es sich beim linken Teil des
    	// Ausdrucks, also bei n[transponiert]*M invers, um den (Zeilen-)Vektor handelt,
    	// welcher im rechten Winkel zu t[M] steht; wichtig ist allerdings: Wenn n[N] und t[M] zueinander
    	// im rechten Winkel stehen, muss nicht n[N]*t[M]=0, sondern n[N][transponiert]*t[M]=0
    	// gelten! D.h. wenn (n[transponiert]*M invers)*t[M]=0 also n[N][transponiert]*t[M]=0
    	// entspricht, dann muss es sich logischerweise bei (n[transponiert]*M invers) um
    	// n[N][transponiert] handeln!
    	// Wird die transponierte Version von n[N] nun noch einmal transponiert, so kommt
    	// wieder das urspruengliche n[N] heraus (zuerst wurden Zeilen und Spalten vertauscht,
    	// nun ein weiteres Mal, sodass die Werte wieder ihre urspruenglichen Positionen
    	// einnehmen). 
    	// Dabei gilt (siehe ECG-Artikel):
    	// n[N]=(n[transponiert]*m invers)[transponiert]= (M invers)[transponiert]*n
    	// Das heit, die Matrix N =(M invers)[transponiert].
    	// Nun wird weiter "getrickst": Wie im ECG-Artikel erwaehnt gilt, dass die inverse Version
    	// einer Matrix A der transponierten Cofaktor-Matrix (einer Matrix, in welcher
    	// fuer jedes Element von A, also fuer jedes aij ein Cofaktor(aij) enthalten ist) von A, dividiert
    	// durch die Determinante von A, entspricht. Da die Laenge des Normalvektor aktuell
    	// aber egal und nur die Orientierung relevant ist (gluecklicherweise, da die 
    	// Multiplikation mit N die Laenge des urspruenglichen n keineswegs bewahren muss), 
    	// kann die Division durch die Determinante einfach weggelassen werden (dies veraendert
    	// nicht die Orientierung).
    	// Kurz:
    	// A invers = (Cofaktormatrix von A)[transponiert]
    	// Was aber, wenn ich nun statt Matrix "A invers" eben N einsetze - oder besser gesagt, wenn ich
    	// (M invers) [transponiert] einsetze? Dann steht auf der linken Seite der Gleichung nicht
    	// nur eine inverse Matrix, sondern eine Matrix, die erst invertiert und dann
    	// transformiert wurde. Dazu muss natuerlich auch die rechte Seite angepasst werden:
    	// A invers = (Cofaktormatrix von A)[transponiert] --->
    	// (A invers)[transponiert] = (Cofaktormatrix von A)[transponiert])[transponiert] --->
    	// (A invers)[transponiert] = Cofaktormatrix von A
    	// Eine transponierte Matrix zu transponieren bringt jene schliesslich, siehe oben,
    	// wieder in ihren "Originalzustand" vor der ersten Transponierung!
    	// Also N = (M invers)[transponiert] = Cofaktormatrix von M --->
    	// N = Cofaktormatrix von M
    	// Nun muessen nur noch die Cofaktoren aller Elemente der Matrix M errechnet werden,
    	// denn N sieht so aus:
    	// N = ((C11, C12, C13), (C21, C22, C23), (C31, C32, C33))
    	// wobei "Cij" bedeutet "Cofactor von mij", also von dem Element der Matrix M in
    	// Zeile i und Spalte j.
    	// Dass N  eine 3x3-Matrix sein muss, steht ja bereits in der obigen, vorgegebenen
    	// Methodenbeschreibung; wenn N der Cofaktormatrix von M entspricht, welche fuer 
    	// jeden Wert von M einen Cofaktor beinhaltet und also die gleichen Dimensionen
    	// haben muesste, wird einmal davon ausgegangen, dass sowohl N als auch M 3x3-Matrizen
    	// sind (Hier muss allerdings noch etwas bedacht werden - siehe unten).
    	// Der Cofaktor eines Wertes kann leicht errechnet werden: In einer 3x3-Matrix A wie dieser
    	// wird selbige bei der Errechnung von Cij zu einer 2x2-Matrix "degradiert", indem
    	// die i-te Zeile und die j-te Spalte gestrichen werden. Von der uebrig bleibenden
    	// 2x2-Matrix wird dann die Determinante errechnet, um den "Minor" von aij zu errechnen 
    	// (jener nennt sich Mij):
    	// A = ((a11, a12, a13), (a21, a22, a23), (a31, a32, a33))
    	// z.B. M23 = det(((a11, a12), (a31, a32))) = a11*a32 - a31*a12
    	// Dies ist nun fast schon das fertige Ergebnis; folgendermassen wird lediglich noch
    	// das Vorzeichen bestimmt (der Minor wird mit -1 multipliziert, wenn i+j ein 
    	// ungerader Wert wird, um den Cofaktor zu erreichen; ist er gerade, so sind an dieser
    	// Stelle Minor und Cofaktor aequivalent):
    	// Cij = ((-1)^(i+j))*Mij, also in diesem Beispiel
    	// C23 = ((-1)^(2+3))*Mij = -1*Mij = -1 * (a11*a32 - a31*a12) = -a11*a32 + a31*a12 =
    	// a31*a12 - a11*a32
    	// Sollte also alles leicht gehen - wenn die nun oft genannte Matrix M jedoch, wie
    	// eingangs erwaehnt, unserer "modelMatrix" (siehe ganz oben) entspricht, dann handelt es sich bei ihr um keine 3x3-, 
		// sondern um eine 4x4-Matrix.
    	// Fuer die Umwandlung in die gewuenschte Form steht eine entsprechende vorgefertigte  Methode bereit, 
		// welche einfach die vierte Zeile und vierte Spalte abschneidet:
    	
		// Anm.: Ehemals "ModelMatrix3x3" genannt; in m wegen der Krze umbenannt.
    	mat3 m = mat3(modelMatrix);
    // Dieses "Wegwerfen" von Zeile 4 und Spalte 4 ist hierbei kein Problem, da dort
    	// eigentlich nur Werte stehen sollten, welche sich auf eventuelle Translationen
    	// beziehen (siehe CG4-Textblatt); da sich im Zuge einer Translation, also Verschiebung,
    	// die Orientierung eines Koerpers nicht aendert, sind diese Informationen fuer die
    	// Berechnung der Oberflaechennormale bzw. fuer die Suche nach N nicht relevant.
    	
    	// Nun muss nur noch nach den obigen Prinzipien und ueberlegungen N, also Ms 
    	// Cofaktormatrix ausgerechnet werden; hierfuer steht bereits eine fertige Formel
    	// in besagtem Artikel bereit:
    	// N = ((m22*m33-m23*m32, m23*m31-m21*m33, m21*m32-m22*m31),
    	//      (m13*m32-m12*m33, m11*m33-m13*m31, m12*m31-m11*m32),
    	//      (m12*m23-m13*m22, m13*m21-m11*m23, m11*m22-m12*m21))
    	// Freilich darf nicht vergessen werden, hier bei 0 zu beginnen anstatt bei 1:
       	// N = ((m11*m22-m12*m21, m12*m20-m10*m22, m10*m21-m11*m20),
    	//      (m02*m21-m01*m22, m00*m22-m02*m20, m01*m20-m00*m21),
    	//      (m01*m12-m02*m11, m02*m10-m00*m12, m00*m11-m01*m10))
    	
    	// ACHTUNG: Siehe http://en.wikibooks.org/wiki/GLSL_Programming/Vector_and_Matrix_Operations - in den mat3-Konstruktor
		// kommen nicht die Elemente in der Reihenfolge "Zeile 0 / Spalte 0, Zeile 0 / Spalte 1, Zeile 0 / Spalte 2, Zeile 1 / Spalte 0 etc."
		// sondern in der Reihenfolge "Zeile 0 / Spalte 0, Zeile 1 / Spalte 0, Zeile 2 / Spalte 0, Zeile 0 / Spalte 1 etc."; sprich,
		// es wird nicht Zeile fr Zeile, sondern Spalte fr Spalte durchgegangen. Somit muss die folgende - auskommentierte - 
		// Gestalt...

    	// mat3 normalMatrix = mat3(	m[1][1]*m[2][2]-m[1][2]*m[2][1], m[1][2]*m[2][0]-m[1][0]*m[2][2], m[1][0]*m[2][1]-m[1][1]*m[2][0],
    	//     							m[0][2]*m[2][1]-m[0][1]*m[2][2], m[0][0]*m[2][2]-m[0][2]*m[2][0], m[0][1]*m[2][0]-m[0][0]*m[2][1],
    	//    							m[0][1]*m[1][2]-m[0][2]*m[1][1], m[0][2]*m[1][0]-m[0][0]*m[1][2], m[0][0]*m[1][1]-m[0][1]*m[1][0]);
    	
		// ...tatschlich derartige angeschrieben werden, damit sie korrekt funktioniert:
		//mat3 normalMatrix = mat3(	m[1][1]*m[2][2]-m[2][1]*m[1][2], m[2][1]*m[0][2]-m[0][1]*m[2][2], m[0][1]*m[1][2]-m[1][1]*m[0][2],
    	//    	                    m[2][0]*m[1][2]-m[1][0]*m[2][2], m[0][0]*m[2][2]-m[2][0]*m[0][2], m[1][0]*m[0][2]-m[0][0]*m[1][2],
    	//    	                    m[1][0]*m[2][1]-m[2][0]*m[1][1], m[2][0]*m[0][1]-m[0][0]*m[2][1], m[0][0]*m[1][1]-m[1][0]*m[0][1]);

		// richtigerweise muss in Opengl in der ersten Zeile die erste Spalte stehen, in der zweiten Zeile die zweite Spalte usw.
		mat3 normalMatrix = mat3(	m[1][1]*m[2][2]-m[2][1]*m[1][2], m[2][0]*m[1][2]-m[1][0]*m[2][2], m[1][0]*m[2][1]-m[2][0]*m[1][1],
									m[2][1]*m[0][2]-m[0][1]*m[2][2], m[0][0]*m[2][2]-m[2][0]*m[0][2], m[2][0]*m[0][1]-m[0][0]*m[2][1],
									m[0][1]*m[1][2]-m[1][1]*m[0][2], m[1][0]*m[0][2]-m[0][0]*m[1][2], m[0][0]*m[1][1]-m[1][0]*m[0][1]);
    // [/Lukas]

	// Nachdem nun endlich die normalMatrix errechnet wurde, kann sie zur Transformation der Normalen in Weltkoordinaten 
	// verwendet werden! Somit wird die anfangs definierte out-Variable "worldNormal" mit dem korrekten Wert "befllt", 
	// nachdem die Normale auch normalisiert wurde.
	worldNormal = normalize(normalMatrix * normal);
    // (4) Errechnung eines ebenfalls fr sptere Berechnungen gebrauchten View-Vektors (verluft von der Oberflchenposition
	// aus zur Kamera):
	// - Im Blinn-Phong-Shader, so wie er hier implementiert ist, wird diese Berechnung erst im fragment shader gettigt.

	// Somit wurden alle ntigen Werte ausgegeben, damit sie - nach ihrer Interpolation - im fragment shader verwendet
	// werden knnen.



	//clipping after definition of gl_Position

	// position of vertex in viewspace
	//vec4 vsPos = modelMatrix * vec4(position, 1);

	//gl_ClipDistance[0] = dot(u_plane0,vsPos);
}