#version 330

uniform mat4 reflectionProjectionModelView;
uniform sampler2D reflectionTexture;
uniform sampler2D normalMap;
uniform vec3 cameraPosition;
uniform float time;
uniform vec3 lichtposition;

in vec3 interpolatedVertexObject;
in vec3 worldPos;
in vec2 fragmentUV;

void main()
{
	vec4 tangent = vec4(1.0, 0.0, 0.0, 0.0);
	vec4 norm = vec4(0.0, 1.0, 0.0, 0.0);
	vec4 binormal = vec4(0.0, 0.0, 1.0, 0.0);

	vec4 vClipReflection = reflectionProjectionModelView * vec4(interpolatedVertexObject.x, 0.0, interpolatedVertexObject.z , 1.0);
	vec2 vDeviceReflection = vClipReflection.xz / vClipReflection.y ;
	vec2 vTextureReflection = vec2(0.5, 0.5) + 0.5 * vDeviceReflection;

	// Framebuffer reflection can have alpha > 1
	//reflectionTextureColor.a = 0.5;

	//normale aus der normalMap lesen
	//vec3 normale = vec3(0.0, 1.0, 0.0);
	vec4 normale = texture2D (normalMap, fragmentUV);
	normale = normale * 2 - 1;
	normale = normalize(normale);

	vec2 resolution;
	resolution.x = 800;
	resolution.y = 600;

	float ripple_strength = 0;

	vec2 texCoord = gl_FragCoord.xy / resolution.xy;
	vec2 offsetTexCoord = texCoord + normale.xz * ripple_strength / gl_FragCoord.z;

	//wellen
	vec2 cPos = -1.0 + 2.0 * gl_FragCoord.xy / resolution.xy;
	float cLength = length(cPos);

	//vec2 uv = gl_FragCoord.xy/resolution.xy+(cPos/cLength)*cos(cLength*12.0-time*4.0)*0.03;
	//normale = (cPos/cLength)*cos(cLength*3.0-time*2.0)*0.03;
	//norm.xz = (cPos/cLength)*cos(cLength*3.0-time*2.0)*0.03;
	norm = normalize(norm);
	vec2 uv = gl_FragCoord.xy/resolution.xy+(cPos/(cLength*2))*cos(cLength*3.0-time*2.0)*0.03;

	vec4 reflectionTextureColor = texture2D (reflectionTexture, uv);//gl_FragCoord.xy / resolution.xy);

	//fresnel = 1 - normale dot einfallenderBlickvector
	float Rmin = 0.1;
	vec3 einfallenderBlickvector = normalize(worldPos - cameraPosition);
	float fresnel = dot(norm.xyz, einfallenderBlickvector);
	if(fresnel < 0)
	{
		fresnel = 0;
	}
	reflectionTextureColor.a = 1 - fresnel * 1.3 - 0.2;// - 0.5;//Rmin + (1 - Rmin) * (1 - fresnel) * (1 - fresnel)* (1 - fresnel)* (1 - fresnel)* (1 - fresnel);

	//specular highlight
	vec3 richtungsvektor = normalize(lichtposition - worldPos);
	float cosTheta = dot(norm.xyz, richtungsvektor);
	if (cosTheta < 0) {
		cosTheta = 0;
    }

	vec3 specular = vec3(1.0, 1.0, 1.0);
	float shininess = 1;
	vec3 lightIntensitySpecular = vec3(1.0, 1.0, 1.0);

	vec3 halbierende = normalize(richtungsvektor + einfallenderBlickvector);

	float cosSigma = dot(norm.xyz, halbierende);
	if (cosSigma < 0) {
		cosSigma = 0;
    }

	//reflectionTextureColor.r += specular.x * cosTheta * lightIntensitySpecular.x * 
    //			pow(cosSigma, shininess);
	//reflectionTextureColor.g += specular.y * cosTheta * lightIntensitySpecular.y * 
    //			pow(cosSigma, shininess);
	//reflectionTextureColor.b += specular.z * cosTheta * lightIntensitySpecular.z * 
    //			pow(cosSigma, shininess);


	gl_FragColor = reflectionTextureColor;

	//gl_FragColor = vec4(1.0f, 1.0f, 1.0f, 1.0f);
}